#!/usr/bin/perl -w
# (c) Copyright 2005-2006. CodeWeavers, Inc.
package CXBottle;
use strict;
use File::Basename;
use File::Path;

use CXLog;
use CXUtils;



sub managed_bottle_dir()
{
    return "$ENV{CX_ROOT}/support";
}

sub set_environment($$)
{
    my ($cxconfig, $msg)=@_;
    my $userenv = $cxconfig->get_section("EnvironmentVariables");
    if (defined $userenv)
    {
        my $traced;
        my $temp = $userenv->get_field_list();
        foreach my $var (@$temp)
        {
            if (!$traced)
            {
                cxlog("$msg:\n");
                $traced=1;
            }
            if ($var =~ /^(CX_BOTTLE|WINEPREFIX)$/)
            {
                cxerr("setting '$var' in an [EnvironmentVariables] section is not allowed\n");
                exit 1;
            }
            my $value=$userenv->get($var);
            $ENV{$var}=expand_string($value);
            if ($var eq "CX_BOTTLE_DIR" and $ENV{$var} !~ m%^/%)
            {
                cxerr("'CX_BOTTLE_DIR' must be an absolute path\n");
                exit 1;
            }
            cxlog(" $var -> $ENV{$var}\n");
        }
    }
}

# Reads all the bottle-independent configuration files
# and sets up the corresponding environment variables.
sub get_crossover_config()
{
    use CXConfig;
    my $productid=CXUtils::get_product_id();

    my $cx_bottle_dir=($ENV{HOME} ? "$ENV{HOME}/.$productid" :
                                    "$ENV{CX_ROOT}/support");

    # Set to the default value so it can be used in [EnvironmentVariables]
    $ENV{CX_BOTTLE_DIR} ||= $cx_bottle_dir;
    my $cxconfig=CXConfig->new("$ENV{CX_ROOT}/etc/$productid.conf",
                               "$cx_bottle_dir/$productid.conf");
    set_environment($cxconfig, "General/User environment variables");
    if ($ENV{CX_BOTTLE_DIR} ne $cx_bottle_dir)
    {
        $cx_bottle_dir=$ENV{CX_BOTTLE_DIR};
        # Also look for a configuration file in $CX_BOTTLE_DIR
        # so it can completely replace ~/.$productid
        $cxconfig->read("$ENV{CX_BOTTLE_DIR}/$productid.conf");
        set_environment($cxconfig, "CX_BOTTLE_DIR environment variables");
        if ($ENV{CX_BOTTLE_DIR} ne $cx_bottle_dir)
        {
            cxerr("setting 'CX_BOTTLE_DIR' in '\$CX_BOTTLE_DIR/$productid.conf' is not allowed\n");
            exit 1;
        }
    }
    return $cxconfig;
}

# Given a bottle name computes the corresponding wineprefix.
# This does not check that the resulting wineprefix is valid.
sub compute_wineprefix($$)
{
    my ($bottle, $scope)=@_;
    my $wineprefix;
    if ($bottle =~ m%^/%)
    {
        $wineprefix=$bottle;
    }
    elsif ($bottle =~ m%/%)
    {
        require Cwd;
        $wineprefix=Cwd::cwd() . "/$bottle";
    }
    elsif (($scope || "private") eq "private")
    {
        $wineprefix="$ENV{CX_BOTTLE_DIR}/$bottle";
        if (!-d $wineprefix)
        {
            my $cx_root_prefix = managed_bottle_dir() . "/$bottle";
            if (-d $cx_root_prefix and -w $cx_root_prefix)
            {
                $wineprefix=$cx_root_prefix;
            }
        }
    }
    else
    {
        $wineprefix = managed_bottle_dir() . "/$bottle";
    }
    return $wineprefix;
}

# Check to see if the given bottle is a symlink, and if so, return
# the referenced bottle
sub get_real_bottle($)
{
    if ($ENV{CX_BOTTLE} eq "default")
    {
        my $wineprefix = get_wineprefix(@_);

        if (defined $wineprefix and -l $wineprefix)
        {
            my $linkeddir = readlink($wineprefix);

            if ($linkeddir =~ m%^/%)
            {
                return $linkeddir;
            }
            elsif ($linkeddir =~ m%/%)
            {
                return dirname($wineprefix) . "/$linkeddir";
            }
            else
            {
                return $linkeddir;
            }
        }
    }

    return $ENV{CX_BOTTLE};
}

sub is_initialized()
{
    return ($ENV{CX_INITIALIZED} || "") eq "$>:$ENV{CX_BOTTLE}";
}

sub get_wineprefix($)
{
    my $wineprefix;
    if (!is_initialized())
    {
        my ($scope)=@_;
        $wineprefix=compute_wineprefix($ENV{CX_BOTTLE}, $scope);
    }
    else
    {
        $wineprefix=$ENV{WINEPREFIX};
    }

    if (-e $wineprefix)
    {
        if (!-d $wineprefix)
        {
            $@="'$wineprefix' is not a directory";
            $wineprefix=undef;
        }
        elsif (!-o $wineprefix)
        {
            $@="'$wineprefix' does not belong to you";
            $wineprefix=undef;
        }
        elsif (!-w $wineprefix)
        {
            $@="'$wineprefix' is not writable";
            $wineprefix=undef;
        }
    }
    return $wineprefix;
}

sub get_bottle_mode($)
{
    my ($cxconfig)=@_;
    my $updater=$cxconfig->get("Bottle","Updater");
    return "private" if (!$updater);

    # Get the cxbottle.conf filename, this should be the last one we read
    my $filename=$cxconfig->get_filenames()->[-1];
    my @st=lstat($filename);
    return ($st[4] == 0) ? "managed" : "stub";
}

sub get_template_directory($)
{
    my ($cxconfig)=@_;
    my $template=$cxconfig->get("Bottle","Template") || "win98";
    my $dir=$template;
    $dir="$ENV{CX_ROOT}/share/crossover/bottle_templates/$dir" if ($dir !~ m%^/$%);
    return (-d $dir ? $dir : undef);
}

sub get_bottle_tag($)
{
    my ($cxconfig)=@_;
    my $productid=CXUtils::get_product_id();
    my $bottleid=$cxconfig->get("Bottle","BottleID");
    return "$productid-$bottleid";
}

sub get_bottle_config($$)
{
    my ($cxconfig, $wineprefix)=@_;

    if (!is_initialized())
    {
        # We must synchronize with other processes to make sure
        # we don't use a half-created bottle.
        my $lockdir=$ENV{TMPDIR} || "/tmp";
        $lockdir.= "/.wine-$>";
        my $err;
        my $productid=CXUtils::get_product_id();
        my $lock=CXUtils::cxlock($lockdir, "$productid-wine.lock", $err);
        if (!defined $lock)
        {
            cxwarn("$err\n");
            cxwarn("locking failed, continuing without a lock\n");
        }

        my $cx_bottle_dir=$ENV{CX_BOTTLE_DIR};
        $cxconfig->read("$wineprefix/cxbottle.conf");
        if ($ENV{CX_BOTTLE_DIR} ne $cx_bottle_dir)
        {
            cxerr("setting 'CX_BOTTLE_DIR' in '\$CX_BOTTLE_DIR/cxbottle.conf' is not allowed\n");
            exit 1;
        }

        if (defined $lock and !CXUtils::cxunlock($lock))
        {
            cxwarn("unable to release the lock\n");
        }
    }
    else
    {
        $cxconfig->read("$wineprefix/cxbottle.conf");
    }
}

# Make the current bottle the default
sub make_default($)
{
    my ($scope)=@_;

    my $wineprefix=get_wineprefix($scope);
    return 0 if (!defined $wineprefix);

    my $bottle_dir=($scope eq "managed") ? managed_bottle_dir() : "$ENV{CX_BOTTLE_DIR}";
    if (!-d $bottle_dir and !mkdir $bottle_dir)
    {
        cxwarn("unable to create the 'default' bottle symlink: $!\n");
        return 0;
    }

    my $default_link="$bottle_dir/default";
    if (-l $default_link and !unlink $default_link)
    {
        cxwarn("unable to delete '$default_link': $!\n");
        return 0;
    }

    # Create a relative symbolic link if possible
    $wineprefix=basename($wineprefix) if (dirname($wineprefix) eq $bottle_dir);

    cxlog("Symlinking '$default_link' to '$wineprefix'\n");
    if (!symlink($wineprefix, $default_link))
    {
        cxwarn("unable to create the 'default' bottle symlink: $!\n");
        return 0;
    }
    return 1;
}

# Returns 1 for success, 0 for failure.
sub create_drive($$$)
{
    my ($wineprefix, $letter, $path)=@_;

    my $linkpath;
    # We check the length of 'letter' because 'z'++ is 'aa'.
    while (length($letter)<=1)
    {
        $linkpath="$wineprefix/dosdevices/$letter:";
        if (! -l $linkpath and ! -e $linkpath)
        {
            # We found an unassigned drive letter.  Time to move on.
            last;
        }
        my $linked=readlink $linkpath;
        if ($linked eq $path)
        {
            # This drive is already set up.  What luck!
            return 1;
        }

        $letter++;
    }

    if (length($letter) > 1)
    {
        cxwarn("Ran out of drive letters for '$path'\n");
        return 0;
    }

    cxlog("Creating drive $letter: -> $path\n");
    if (!symlink($path, $linkpath))
    {
        cxwarn("unable to create drive letter ($letter) symlink: $!\n");
    }
    return 1;
}

sub get_device_link($)
{
    my ($link)=@_;

    my $path = readlink $link;
    while (defined $path and $path ne "")
    {
        $path =~ s/(..\/)*//;
        $link = "/dev/$path";

        $path = readlink $link;
    }

    return $link;
}

sub setup_drives($)
{
    my ($wineprefix)=@_;

    if (-d "$wineprefix/dosdevices")
    {
        # Delete broken drives from previous installs, and drives that are
        # no longer valid because the bottle was moved.
        for (my $i=0; $i<26; $i++)
        {
            my $letter=chr(ord('a')+$i);
            my $linkpath="$wineprefix/dosdevices/$letter:";
            if (!-l $linkpath)
            {
                cxwarn("drive $letter: should be a symbolic link\n") if (-e _);
            }
            elsif (!-d $linkpath)
            {
                cxlog("drive $letter: does not point to a directory. Deleting it\n");
                if (!unlink $linkpath)
                {
                    cxwarn("unable to delete '$linkpath': $!\n");
                }
            }
        }
    }
    elsif (!mkdir("$wineprefix/dosdevices", 0777))
    {
        cxerr("unable to create the '$wineprefix/dosdevices' directory: $!\n");
        cxerr("drive creation cancelled\n");
        return;
    }

    create_drive($wineprefix, 'c', "../drive_c");
    # Force the re-creation of the y: drive
    my $home=$ENV{HOME} || (getpwuid($>))[7];
    unlink "$wineprefix/dosdevices/y:";
    create_drive($wineprefix, 'y', $home);
    create_drive($wineprefix, 'z', "/");

    # FIXME: Switch to hashtables to store the cdrom and floppy lists
    my ($fstab_cdroms, $fstab_floppies) = CXUtils::parse_fstab("/etc/fstab");
    my ($am_cdroms, $am_floppies) = CXUtils::parse_automaster("/etc/auto.master");

    my @floppies;
    foreach my $floppy (@$fstab_floppies, @$am_floppies)
    {
        my $floppylink=get_device_link($$floppy[1]);

        # Move on if we already have a letter for this one.
        next if ((grep /$floppylink/,@floppies)[0]);
        next if (!-d $$floppy[0]);
        push @floppies, $floppylink;

        # Always start from the same letter to correctly deal with
        # pre-existing drives.
        last if (!create_drive($wineprefix, "a", $$floppy[0]));
    }

    my @cdroms;
    foreach my $cdrom (@$fstab_cdroms, @$am_cdroms)
    {
        # Don't create drive letters for loop devices
        next if (@$cdrom[4]);

        my $cdromlink=get_device_link($$cdrom[1]);

        # Move on if we already have a letter for this one.
        next if ((grep /$cdromlink/,@cdroms)[0]);
        next if (!-d $$cdrom[0]);
        push @cdroms, $cdromlink;

        # Always start from the same letter to correctly deal with
        # pre-existing drives.
        last if (!create_drive($wineprefix, "m", $$cdrom[0]));
    }
}

sub update_bottle($$)
{
    if (!is_initialized())
    {
        my ($cxconfig, $wineprefix)=@_;
        my $update;
        if (get_bottle_mode($cxconfig) ne "stub")
        {
            my $template_dir=get_template_directory($cxconfig);
            if (!defined $template_dir)
            {
                cxerr("unsupported template type for bottle '$ENV{CX_BOTTLE}'\n");
                return 0;
            }
            cxlog("template_dir is not defined!\n") if (!defined $template_dir);
            my $mtime_ref=(stat("$template_dir/setup"))[9];
            my $mtime_bottle=(stat("$wineprefix/cxbottle.conf"))[9];
            if (!$mtime_ref or !$mtime_bottle or $mtime_bottle <= $mtime_ref)
            {
                $update=1;
            }
        }
        else
        {
            $update=1;
        }
        if ($update)
        {
            # FIXME: We can't check the return code. Plus if things go wrong
            # it becomes interactive :-(
            cxsystem("$ENV{CX_ROOT}/bin/wine","--bottle",$ENV{CX_BOTTLE},"--ux-app","true");
        }
    }
    return 1;
}

sub run_scripts($$)
{
    my ($dir, $args)=@_;

    local *DIR;
    if (!opendir(DIR, $dir))
    {
        cxlog("unable to open the '$dir' directory: $!\n");
        return 0;
    }
    my @dentries=readdir DIR;
    closedir(DIR);
    foreach my $dentry (sort @dentries)
    {
        next if ($dentry !~ /^[0-9][0-9]\.[^.~]+$/);
        next if (!-x "$dir/$dentry" or !-f _);
        cxsystem("$dir/$dentry", @$args);
    }
}

sub run_bottle_hooks($)
{
    my ($args)=@_;
    foreach my $dir ("$ENV{CX_ROOT}/support/scripts.d",
                     "$ENV{CX_BOTTLE_DIR}/scripts.d",
                     "$ENV{WINEPREFIX}/scripts.d")
    {
        run_scripts($dir, $args) if (-d $dir);
    }
}

sub fix_ownership_permissions($$)
{
    my ($rootdir, $scope)=@_;

    # Fix the user/group ownership, and maybe the permissions too
    my $shared=($scope eq "managed");
    my $egid=$);
    $egid=~s/ .*$//;
    local *DIR;
    my @dirs=($rootdir);
    while (@dirs)
    {
        my $dir=shift @dirs;
        if (!opendir(DIR, $dir))
        {
            cxerr("unable to open the '$dir' directory: $!\n");
            next;
        }
        foreach my $dentry (readdir DIR)
        {
            next if ($dentry =~ /^\.\.?$/);
            $dentry="$dir/$dentry";

            my @st=lstat($dentry);
            push @dirs, $dentry if (-d _);
            if (CXLog::is_on())
            {
                cxlog(sprintf("%04o", $st[2] & 07777),
                      " - $st[4] - $st[5] - $dentry\n");
            }
            if ($st[4] != 0 or $st[5] != $egid)
            {
                cxlog("chown 0:$egid $dentry\n");
                if (-l _)
                {
                    # Perl does not have lchown :-(
                    cxsystem("chown", "-h", "0:$egid", $dentry);
                }
                else
                {
                    chown 0, $egid, $dentry;
                }
            }

            # If the file already belonged to us, then assume it already has
            # the right permissions (but the group ownership is system
            # dependent and needed fixing anyway)
            next if ($st[4] == 0);

            # Don't change the permissions on symbolic links
            next if (-l _);

            my $mode=($st[2] & 07777) & ~umask;
            $mode|= (-d _ ? 0055 : 0044) if ($shared);
            if ($mode != $st[2])
            {
                if (CXLog::is_on())
                {
                    cxlog("chmod ", sprintf("%04o", $mode), " $dentry\n");
                }
                chmod $mode, $dentry;
            }
        }
        closedir(DIR);
    }
}

sub get_desktopdata_dir($$)
{
    my ($scope, $tag)=@_;
    return "$ENV{WINEPREFIX}/desktopdata" if (defined $ENV{WINEPREFIX});
    return undef if (!defined $tag);
    return managed_bottle_dir() . "/desktopdata/$tag" if ($scope eq "managed");
    return "$ENV{CX_BOTTLE_DIR}/desktopdata/$tag";
}

sub removeall_desktopdata_dirs($$)
{
    my ($pattern, $subdir)=@_;
    foreach my $dir (managed_bottle_dir() . "/desktopdata",
                     "$ENV{CX_BOTTLE_DIR}/desktopdata")
    {
        next if (!-d $dir or !-w $dir);
        local *DIR;
        if (!opendir(DIR, $dir))
        {
            cxlog("unable to open the '$dir' directory: $!\n");
            next;
        }
        foreach my $dentry (readdir DIR)
        {
            if ($dentry =~ /^$pattern/)
            {
                my $d="/$dentry/$subdir";
                if (-d "$dir$d")
                {
                    cxlog("Deleting the '$dir$d' directory\n");
                    if (!rmtree "$dir$d")
                    {
                        cxwarn("unable to delete the '$dir$d' directory: $!\n");
                    }
                    CXUtils::garbage_collect_subdirs($dir, dirname($d), 1);
                }
            }
        }
        closedir(DIR);
        cxlog("Deleted the '$dir' directory\n") if (rmdir $dir);
    }
}

return 1;
