#!/usr/bin/perl -w
# (c) Copyright 2005. CodeWeavers, Inc.
package CXMenuShortcut;
use strict;
use File::Path;
use CXLog;
use CXUtils;
use CXMenu;

sub detect($$$$)
{
    my ($class, $cxoptions, $cxconfig, $gui_info)=@_;

    my $prefix=($gui_info->{preferred_scope} eq "managed" ? "Managed" : "Private");
    my $shortcutdirs=$cxconfig->get("CrossOver", "${prefix}ShortcutDirs") || "";

    my $self={
        tag          => $cxoptions->{tag},
        desktopdata  => $cxoptions->{desktopdata},
        shortcutdirs => $shortcutdirs
    };
    bless $self, $class;
    return ($self);
}

sub id($)
{
    return "CXMenuShortcut/";
}

sub get_directories($)
{
    my ($self)=@_;

    my $directories=$self->{directories};
    if (!defined $directories)
    {
        $directories=[];
        foreach my $dir (split /:+/, $self->{shortcutdirs})
        {
            next if ($dir eq "");
            # Skip missing directories without complaining
            next if (!-e $dir);
            if (!-d _)
            {
                cxwarn("'$dir' is not a valid shortcut directory\n");
            }
            elsif (!-w _)
            {
                cxwarn("'$dir' is not writable\n");
            }
            else
            {
                push @$directories, $dir;
            }
        }
        $self->{directories}=$directories;
    }
    return $directories;
}

sub install($$)
{
    my ($self, $components)=@_;

    my $menu=@$components[-1];
    return 1 if ($menu->{shortcut} eq "");
    cxlog("Installing shortcut '$menu->{shortcut}'\n");

    if ($menu->{is_dir})
    {
        cxerr("'$menu->{rawpath}': directories cannot have shortcuts. Ignoring it\n");
        return 0;
    }

    my $root="$self->{desktopdata}/cxmenu/Shortcuts";
    if (!-d $root and !cxmkpath($root))
    {
        cxerr("unable to create '$root': $@\n");
        return -1; # Fatal error
    }

    my $script="$root/$menu->{tag}." . mangle_string($menu->{shortcut});
    cxlog("Creating '$script'\n");
    chmod 0700, $script;
    my $fh=FileHandle->new(">$script");
    if (!$fh)
    {
        cxerr("unable to open '$script' for writing: $!\n");
        return -1; # Fatal error
    }
    print $fh "#!/bin/sh\n";
    print $fh "exec $menu->{command} \"\$\@\"\n";
    $fh->close();
    my $mode=0755 & ~umask();
    chmod $mode, $script;

    foreach my $dir (@{$self->get_directories()})
    {
        my $shortcut="$dir/$menu->{shortcut}";

        # First delete the existing shortcut if any,
        # even if it belongs to another CrossOver product.
        # Note: -l does an lstat() but we want a stat() for -f and -e
        #       so we must be careful when using _
        if (-l $shortcut and (-f $shortcut or !-e _))
        {
            my $dst=readlink $shortcut;
            if ($dst =~ m%/desktopdata/cxmenu/Shortcuts/%)
            {
                cxlog("Deleting '$shortcut'\n");
                unlink $shortcut;
            }
            else
            {
                cxlog("'$shortcut' is not a CrossOver shortcut\n");
            }
        }
        cxlog("Symlinking '$shortcut'\n");
        if (!symlink $script, $shortcut)
        {
            cxlog("unable to symlink '$shortcut' to '$script'\n");
        }
    }
    return 1;
}

sub query($$)
{
    # Don't report the CXMenuShortcut install status
    return ("", "");
}

sub uninstall($$)
{
    my ($self, $components)=@_;

    my $menu=@$components[-1];
    return 1 if (!defined $menu->{shortcut} or $menu->{shortcut} eq "");

    my $script="$self->{desktopdata}/cxmenu/Shortcuts/$menu->{tag}." .
               mangle_string($menu->{shortcut});
    if (-f $script)
    {
        cxlog("Deleting '$script'\n");
        if (!unlink $script)
        {
            cxerr("unable to delete '$script': $!\n");
        }
    }

    foreach my $dir (@{$self->get_directories()})
    {
        my $shortcut="$dir/$menu->{shortcut}";
        # Note: -l does an lstat() but we want a stat() for -f and -e
        #       so we must be careful when using _
        if (-l $shortcut and (-f $shortcut or !-e _))
        {
            my $dst=readlink $shortcut;
            if ($dst =~ m%/desktopdata/cxmenu/Shortcuts/$menu->{tag}\.%)
            {
                cxlog("Deleting '$shortcut'\n");
                if (!unlink $shortcut)
                {
                    cxerr("unable to delete '$shortcut': $!\n");
                }
            }
            else
            {
                cxlog("'$shortcut' is not a CrossOver shortcut\n");
            }
        }
    }

    return 1;
}

sub is_legacy_file($)
{
    my ($filename)=@_;

    my $fh=FileHandle->new($filename);
    return 0 if (!$fh);
    while (my $line=<$fh>)
    {
        if ($line =~ /^#cxmenu auto generated/)
        {
            close($fh);
            return 1;
        }
    }
    close($fh);
    return 0;
}

sub removeall($$)
{
    my ($self, $pattern)=@_;
    local *DIR;

    if ($pattern eq "legacy")
    {
        # The legacy shortcuts were very different :-(
        if (-w "/usr/bin" and opendir(DIR, "/usr/bin"))
        {
            foreach my $dentry (readdir DIR)
            {
                next if ($dentry =~ /^\.\.?$/);
                $dentry="/usr/bin/$dentry";

                # Note: -l does an lstat() but we want a stat() for -f
                #       and -e so we must be careful when using _
                next if (!-l $dentry or (!-f $dentry and -e _));
                my $dst=readlink $dentry;
                if ($dst =~ m%$ENV{CX_ROOT}/bin/% and is_legacy_file($dentry))
                {
                    cxlog("Deleting '$dentry'\n");
                    if (!unlink $dentry)
                    {
                        cxerr("unable to delete '$dentry': $!\n");
                    }
                }
            }
            closedir(DIR);
        }
        if (-w "$ENV{CX_ROOT}/bin" and opendir(DIR, "$ENV{CX_ROOT}/bin"))
        {
            foreach my $dentry (readdir DIR)
            {
                next if ($dentry =~ /^\.\.?$/);
                $dentry="$ENV{CX_ROOT}/bin/$dentry";

                if (is_legacy_file($dentry))
                {
                    cxlog("Deleting '$dentry'\n");
                    if (!unlink $dentry)
                    {
                        cxerr("unable to delete '$dentry': $!\n");
                    }
                }
            }
            closedir(DIR);
        }
        return 1;
    }

    if ($self->{tag} and $self->{tag} =~ /^$pattern/)
    {
        my $dir="$self->{desktopdata}/cxmenu/Shortcuts";
        if (-d $dir)
        {
            cxlog("Deleting the '$dir' directory\n");
            if (!rmtree $dir)
            {
                cxerr("unable to delete the '$dir' directory: $!\n");
            }
        }
        CXUtils::garbage_collect_subdirs($self->{desktopdata}, "/cxmenu", 1);
    }
    else
    {
        CXBottle::removeall_desktopdata_dirs($pattern, "/cxmenu/Shortcuts");
    }

    $pattern =~ s/\$$/\\./;
    foreach my $dir (@{$self->get_directories()})
    {
        next if (!opendir(DIR, $dir));
        foreach my $dentry (readdir DIR)
        {
            next if ($dentry =~ /^\.\.?$/);
            $dentry="$dir/$dentry";

            # Note: -l does an lstat() but we want a stat() for -f
            #       and -e so we must be careful when using _
            next if (!-l $dentry or (!-f $dentry and -e _));
            my $dst=readlink $dentry;
            if ($dst =~ m%/desktopdata/cxmenu/Shortcuts/$pattern%)
            {
                cxlog("Deleting '$dentry'\n");
                if (!unlink $dentry)
                {
                    cxerr("unable to delete '$dentry'\n");
                }
            }
            else
            {
                cxlog("'$dentry' is not a CrossOver shortcut\n");
            }
        }
        closedir(DIR);
    }

    return 1;
}

sub finalize($)
{
    my ($self)=@_;
    if (defined $self->{desktopdata})
    {
        CXUtils::garbage_collect_subdirs($self->{desktopdata}, "/cxmenu/Shortcuts", 1);
    }
    return 1;
}

return 1;
